/********************************************************************
 * RS-232 Computer Interface Board
 *  Copyright (c) 2012 - Vanderhaegen Bart, bvsystems.be
 *
 *
 * This firmware is derived from my RS-232 interface. It now doesn't
 * control shift registers, but I2C ports (PCF8574AP).
 *
 * Any data is validated by a checksum.
 ********************************************************************/

#include <htc.h>;

// set controller options
__CONFIG(FOSC_HS & WDTE_OFF & PWRTE_OFF & MCLRE_ON & BOREN_OFF & LVP_OFF & CP_OFF & CPD_OFF);

// definition of program behavior
#define MAX_DATA_PACKETS 20
#define _XTAL_FREQ 19668000

// other definitions
#define true 	1
#define false 	0
#define MSG_OK 	0x0F
#define MSG_ERR 0xF0

// definition of hardware I/O ports
#define LED_ERR		RB0
#define LED_DATA 	RB7

// definitions of the data packets
char packets[MAX_DATA_PACKETS+1];	// packet
char *packetPntr;					// pointer to current index
char cycleKeeper=0;					// time-out aider variable

// array with the I2C addresses
char i2cAddresses[]={0x70,0x72,0x74,0x76,0x78,0x7A,0x7C,0x7E};

// define I2C communications
#define SCL TRISB5          // serial clock (requires pull-up)
#define SDA TRISB4          // serial data(requires pull-up
#define SCL_IN  RB5          //
#define SDA_IN  RB4          //

// Function prototypes
void interrupt irq(void);
void sendToBuffers(char *ptrDataByte,char numberOfDataBytes);
void clearAllOutputs();
void initData(void);
void initUART(void);
void flashyLED(void);
bit isDataValid(char *data,char numberOfDataBytesToReceive);
void main(void);
void transmitResponse(char returnCode);
void delay(void);
void i2cStart(void);
void i2cStop(void);
bit waitForAck(void);
void clock(void);
void sendData(char Data);
void giveI2Ccommand(char address,char data);
void delay_ms(int ms);



/********************************************************************
 * Interrupt Handler
 *  Puts the received data into the packets buffer, and handles
 *  the time-out event.
 ********************************************************************/
void interrupt irq(void) {

	if (TMR1IF) {
		// timer overflow: data time-out detected
		static char cycleKeeper=0;

		if (cycleKeeper++>5) {
			cycleKeeper=0;
			initData();
		}

		TMR1IF=0;
	}

	if (RCIF) { 
		// data received. First we need to analyze if the 
		if (FERR) {
			// A framing error occured.
			LED_ERR=1;
			FERR=0;
		}
		else if (OERR) {
			// An overrun error occured
			LED_ERR=1;
			CREN=0;
			CREN=1;
		}
		else {
			// The received byte was OK
			cycleKeeper=0;	// reset time-out

			while(RCIF) {
				// read the RCREG as long as the RCIF is "1".
				*packetPntr=RCREG;
	
				// Set output LEDs
				LED_DATA=!LED_DATA;
			
				if ((packetPntr-packets)<MAX_DATA_PACKETS-1) {
					packetPntr++;
				} // end if ((packetPntr-packets)<MAX_DATA_PACKETS-1)
			} // end while(RCIF)
		} // end if (FERR==1)
	} // end if (RCIF)
}




/********************************************************************
 * K8000 Output controllers
 *  These functions will control the outputs of the K8000.
 ********************************************************************/
// send a certain number of databytes to the I2C ports
void sendToBuffers(char *ptrDataByte,char numberOfDataBytes) {

	char *adrPntr=i2cAddresses;

	// first make sure only 8 bytes are processed
	if (numberOfDataBytes>8) numberOfDataBytes=8;

	// loop through all received bytes. Convert to I2C commands
	for (char byteCnt=0;byteCnt<numberOfDataBytes;byteCnt++) {
		giveI2Ccommand(*adrPntr,~(*ptrDataByte));		
		ptrDataByte++;
		adrPntr++;
	}
}


// clear all the ouptuts
void clearAllOutputs() {
	char *adrPntr=i2cAddresses;
	for (char i=0;i<sizeof(i2cAddresses);i++) {
		giveI2Ccommand(*adrPntr,0xFF);
		adrPntr++;
	}
}





/**************************************************************************
 * Implementation of the I2C protocol
 *************************************************************************/

  // initiate a new I2C command
  void giveI2Ccommand(char address,char data) {
    i2cStart();   
    sendData(address);
    waitForAck();
    sendData(data);
    waitForAck();
    i2cStop();
  }


  // send 8 bits of data over the I2C bus
  // Assumption: SCL is already low
  void sendData(char Data) {
    unsigned char i;
  
    for (i=8;i>0;i--) {
      if(Data & 0x80) {
        SDA=1;
      }
      else {
        SDA_IN=0; SDA=0;
      }

      SCL=1;
      delay();
      Data<<=1;
      SCL_IN=0; SCL=0;
      delay();
    }
  }


  // send the I2C start sequence
  // SDA must go from high to low while SCL is high.
  void i2cStart() {
    SDA=1;    
    delay();
    SCL=1;    
    delay();  
    SDA_IN=0; SDA=0;       
    delay();
    SCL_IN=0; SCL=0; 
    delay();
  }

  
  // send the I2C stop sequence
  // SDA must go from low to high while SCL is high
  void i2cStop() {
    SDA_IN=0; SDA=0;    
    delay();  
    SCL=1;    
    delay();    
    SDA=1;    
    delay();
  }



  // wait for an acknowledge signal
  // Receiver must make SDA low while SCL is high
  // The waitForAck times out after approximately 10 ms

  bit waitForAck() {
    SDA=1;      
    delay();    
    SCL=1;  
    unsigned int timeout=1923;
    while (SDA_IN && (--timeout > 0));
    SCL_IN=0; SCL=0;    
    delay();
    return (timeout>0);
  }



  // send the clock pulse
  // SCL must be made high for a short period of time
  void clock() {
    delay();  
    SCL=1;    
    delay();    
    SCL_IN=0; SCL=0;    
    delay();  
  }


  // send the acknowledge pulse required by the slave device
  void sendAck() {
    SDA_IN=0; 
    SDA=0;    
    delay();
    SCL=1;      
    delay();    
    SCL_IN=0; SCL=0;  
    delay();    
  }


  // delay execution
  void delay(void){
    __delay_us(100);
  }




/********************************************************************
 * Data Handling
 *   This code is written for a 9600 baud communication
 ********************************************************************/
// Reset the incoming databuffer
void initData(void) {
	// clearing the data is not necessary
	packetPntr=packets;
}


// This function will validate our data
bit isDataValid(char *data,char numberOfDataBytesToReceive) {
	if (*data==0xF0) {
		// valid start byte
		int checksum=*data;

		for (char i=0;i<numberOfDataBytesToReceive+1;i++) {
			data++;
			checksum+=*data;
		}

		data++;

		return ((char)(checksum%256)==*data);
	}
	else {
		// incorrect startbyte
		return false;
	}
}





/********************************************************************
 * Communication Initialisation
 *   This code is written for a 9600 baud communication
 ********************************************************************/
void initUART(void) {
	SPBRG=31; 	// baud rate generator register
	BRGH=0;		// Enable high baud-rate

	// Enable asynchronous communication
	SYNC=0;		// Asynchronous mode
	SPEN=1;		// Enable serial port
	TRISB1=1;	// Required by UART (datasheet)
	TRISB2=1;	// Required by UART (datasheet)

	// Enable 8-bit data words
	TX9=0;
	RX9=0;

	// Enable Interrupts
	RCIE=1;  	// USART interrupt
	CREN=1;  	// Continuous recieve enable bit

}


// this code will transmit a confirmation code
void transmitResponse(char returnCode) {
	TXREG=returnCode;		// set byte
	TXEN=1;					// enable transmission
	while(TXIF==0);			// 
	//TXEN=0;				// Todo: Transmission fails when the TXEN is
							// disabled again?
}



/********************************************************************
 * LED output controllers
 *   These functions will control the output LEDs
 ********************************************************************/
// flash the DATA LED 25 times
void flashyLED(void) {
	for (char i=0;i<50;i++) {
		__delay_ms(50);
		LED_DATA=!LED_DATA;
	}

	LED_DATA=0;
}





/********************************************
 * The main sub of the program
 *
 *
 *******************************************/
void main(void) {

	// define the PIC's ports
	TRISA=0x00;
	TRISB=0b00000110;

	// clear the output of the K8000
	clearAllOutputs();

	// set timeout timer properties
	T1CKPS1=1;		// prescaler 1:8
	T1CKPS0=1;		// prescaler 1:8
	T1OSCEN=1;		// enable oscillator
	TMR1CS=0;		// use internal clock
	TMR1ON=1;		// enable timer
	TMR1IE=1;		// enable timer interrupt

	// clear the ports
	PORTA=0x00;
	PORTB=0x00;

	// initialize the data
	initData();

	// setup the RS-232 system
	initUART();

	// test our system
	flashyLED();

	// enable the interrupts
	PEIE=1;  	// Unmasked periheral interrupts (timer/USART)
	ei();

	// here our actual programloop will begin
	while (true) {
		char *receiveBuffer=packets;	// pointer to buffer

		if ((packetPntr-receiveBuffer)>1) {
			// number of bytes is received			
			receiveBuffer++;
			char numberOfDataBytesToReceive=*receiveBuffer;

			if ((packetPntr-receiveBuffer)==numberOfDataBytesToReceive+2) {
				// we have what we needed
				receiveBuffer++;
				if(isDataValid(packets,numberOfDataBytesToReceive)) {
					// the data is valid
					sendToBuffers(receiveBuffer,numberOfDataBytesToReceive);
					LED_ERR=0;
					transmitResponse(MSG_OK);
				}
				else {
					// the data is invalid
					LED_ERR=1;
					transmitResponse(MSG_ERR);
				}

				// reinitialize the data
				initData();
			}
		} // end if ((packetPntr-receiveBuffer)>1)
	} // while (true)
}
